from typing import List, Optional, Set, Tuple

from django.apps import apps
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations import Migration
from django.db.migrations.graph import Node
from django.db.migrations.loader import MigrationLoader

from django_test_migrations.exceptions import MigrationNotInPlan
from django_test_migrations.types import MigrationPlan, MigrationTarget


def all_migrations(
    database: str = DEFAULT_DB_ALIAS,
    app_names: Optional[List[str]] = None,
) -> List[Node]:
    """
    Returns the sorted list of migrations nodes.

    The order is the same as when migrations are applied.

    When you might need this function?
    When you are testing the migration order.

    For example, imagine that you have a direct dependency:
    ``main_app.0002_migration`` and ``other_app.0001_initial``
    where ``other_app.0001_initial`` relies on the model or field
    introduced in ``main_app.0002_migration``.

    You can use ``dependencies`` field
    to ensure that everything works correctly.

    But, sometimes migrations are squashed,
    sometimes they are renamed, refactored, and moved.

    It would be better to have a test that will ensure
    that ``other_app.0001_initial`` comes after ``main_app.0002_migration``.
    And everything works as expected.
    """
    loader = MigrationLoader(connections[database])

    if app_names:
        _validate_app_names(app_names)
        targets = [
            key
            for key in loader.graph.leaf_nodes()
            if key[0] in app_names
        ]
    else:
        targets = loader.graph.leaf_nodes()
    return _generate_plan(targets, loader)


def nodes_to_tuples(nodes: List[Node]) -> List[Tuple[str, str]]:
    """Utility function to transform nodes to tuples."""
    return [
        (node[0], node[1])
        for node in nodes
    ]


def _validate_app_names(app_names: List[str]) -> None:
    """
    Validates the provided app names.

    Raises ```LookupError`` when incorrect app names are provided.
    """
    for app_name in app_names:
        apps.get_app_config(app_name)


def _generate_plan(
    targets: List[Node],
    loader: MigrationLoader,
) -> List[Node]:
    plan = []
    seen: Set[Node] = set()

    # Generate the plan
    for target in targets:
        for migration in loader.graph.forwards_plan(target):
            if migration not in seen:
                node = loader.graph.node_map[migration]
                plan.append(node)
                seen.add(migration)
    return plan


def truncate_plan(
    targets: List[MigrationTarget],
    plan: MigrationPlan,
) -> MigrationPlan:
    """Truncate migrations ``plan`` up to ``targets``.

    This method is used mainly to truncate full/clean migrations plan
    to get as broad plan as possible.
    By "broad" plan we mean plan with all targets migrations included
    as well as all older migrations not related with targets.
    "Broad" plan is needed mostly to make ``Migrator`` API developers'
    friendly, just to not force developers to include migrations targets
    for all other models they want to use in test (e.g. to setup some
    model instances) in ``migrate_from``.
    Such plan will also produce database state really similar to state
    from our production environment just before new migrations are applied.
    Migrations plan for targets generated by Django's
    ``MigrationExecutor.migration_plan`` is minimum plan needed to apply
    targets migrations, it includes only migrations targets with all its
    dependencies, so it doesn't fit to our approach, that's why following
    function is needed.

    """
    if not targets or not plan:
        return plan

    target_max_index = max(_get_index(target, plan) for target in targets)
    return plan[:(target_max_index + 1)]


def _get_index(target: MigrationTarget, plan: MigrationPlan) -> int:
    try:
        index = next(
            index
            for index, (migration, _) in enumerate(plan)  # noqa: WPS405, WPS414
            if _filter_predicate(target, migration)
        )
    except StopIteration:
        raise MigrationNotInPlan(target)
    return index - (target[1] is None)


def _filter_predicate(target: MigrationTarget, migration: Migration) -> bool:
    # when ``None`` passed as migration name then initial migration from
    # target's app will be chosen and handled properly in ``_get_index``
    # so in final all target app migrations will be excluded from plan
    index = 2 - (target[1] is None)
    return (migration.app_label, migration.name)[:index] == target[:index]



from typing_extensions import final

from django_test_migrations.db.backends.base.configuration import (
    BaseDatabaseConfiguration,
)
from django_test_migrations.types import DatabaseSettingValue


@final
class DatabaseConfiguration(BaseDatabaseConfiguration):
    """Interact with PostgreSQL database configuration."""

    vendor = 'postgresql'
    statement_timeout = 'statement_timeout'

    def get_setting_value(self, name: str) -> DatabaseSettingValue:
        """Retrieve value of PostgreSQL database's setting with ``name``."""
        with self.connection.cursor() as cursor:
            cursor.execute(
                (
                    'SELECT setting FROM pg_settings ' +  # noqa: S608
                    'WHERE name = %s;'  # noqa: WPS323
                ),
                (name,),
            )
            setting_value = cursor.fetchone()
            if not setting_value:
                return super().get_setting_value(name)
            return setting_value[0]



from fnmatch import fnmatch
from typing import FrozenSet, List, Tuple

from django.conf import settings
from django.core.checks import CheckMessage, Warning
from typing_extensions import Final

_IgnoreAppSpec = FrozenSet[str]

_IgnoreMigrationSpec = FrozenSet[Tuple[str, str]]

#: We use this type hint to represent ignore rules for migrations.
_IgnoreSpec = Tuple[_IgnoreAppSpec, _IgnoreMigrationSpec]

#: We use this value as a unique identifier of this check.
CHECK_NAME: Final = 'django_test_migrations.checks.autonames'

#: Settings name for this check to ignore some migrations.
_SETTINGS_NAME: Final = 'DTM_IGNORED_MIGRATIONS'

# Special key to ignore all migrations inside an app
_IGNORE_APP_MIGRATIONS_SPECIAL_KEY: Final = '*'


def _is_ignored(
    app_label: str, migration_name: str, ignored: _IgnoreSpec,
) -> bool:
    ignored_apps, ignored_migrations = ignored

    return (
        app_label in ignored_apps or
        (app_label, migration_name) in ignored_migrations
    )


def _build_ignores() -> _IgnoreSpec:
    ignored_migrations: _IgnoreMigrationSpec = getattr(
        settings, _SETTINGS_NAME, frozenset(),
    )

    ignored_apps: _IgnoreAppSpec = frozenset(
        app_label
        for app_label, migration_name in ignored_migrations
        if migration_name == _IGNORE_APP_MIGRATIONS_SPECIAL_KEY
    )

    return ignored_apps, ignored_migrations


def check_migration_names(*args, **kwargs) -> List[CheckMessage]:
    """
    Finds automatic names in available migrations.

    We use nested import here, because some versions of django fails otherwise.
    They do raise:
    ``django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet.``
    """
    from django.db.migrations.loader import MigrationLoader  # noqa: WPS433

    loader = MigrationLoader(None, ignore_no_migrations=True)
    loader.load_disk()

    messages = []
    ignores = _build_ignores()

    for app_label, migration_name in loader.disk_migrations.keys():
        if _is_ignored(app_label, migration_name, ignores):
            continue

        if fnmatch(migration_name, '????_auto_*'):
            messages.append(
                Warning(
                    'Migration {0}.{1} has an automatic name.'.format(
                        app_label, migration_name,
                    ),
                    hint=(
                        'Rename the migration to describe its contents, ' +
                        "or if it's from a third party app, add to " +
                        _SETTINGS_NAME
                    ),
                    id='{0}.W001'.format(CHECK_NAME),
                ),
            )
    return messages


CHECKS: Final = (check_migration_names,)



from django_test_migrations.types import MigrationTarget


class MigrationNotInPlan(Exception):
    """``MigrationTarget`` not found in migrations plan."""

    def __init__(self, migration_target: MigrationTarget) -> None:  # noqa: D107
        self.migration_target = migration_target

    def __str__(self) -> str:
        """String representation of exception's instance."""
        return 'Migration {0} not found in migrations plan'.format(
            self.migration_target,
        )



# -*- coding: utf-8 -*-
# Generated by Django 2.2.7 on 2019-11-19 20:00

from django.db import migrations, models


class Migration(migrations.Migration):
    """Initial migration."""

    initial = True
    dependencies = []

    operations = [
        migrations.CreateModel(
            name='SomeItem',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'string_field',
                    models.CharField(max_length=50),  # noqa: WPS432
                ),
            ],
        ),
    ]



import django
import pytest
from django.apps import apps
from django.core.management import call_command
from django.db import DEFAULT_DB_ALIAS
from django.db.models.signals import post_migrate, pre_migrate
from typing_extensions import Final

from django_test_migrations.signals import mute_migrate_signals

# value for ``dispatch_uid`` is needed to disconnect signal receiver
# registered for testing purposes to which we do not have any reference
# outside of test function
DISPATCH_UID: Final = 'test_migrate_signals'


# Dummy signal receiver function
def _my_callback(sender, **kwargs):
    """Mock that does nothing."""


@pytest.fixture()
def _disconnect_receivers():
    """Disconnect testing receiver of ``pre_migrate`` or ``post_migrate``."""
    yield
    main_app_config = apps.get_app_config('main_app')
    pre_migrate.disconnect(sender=main_app_config, dispatch_uid=DISPATCH_UID)
    post_migrate.disconnect(sender=main_app_config, dispatch_uid=DISPATCH_UID)


@pytest.mark.parametrize('signal', [pre_migrate, post_migrate])
def test_migrate_signal_muted(signal):
    """Ensure the context manager does indeed silences the signals."""
    signal.connect(_my_callback)
    assert signal.receivers
    with mute_migrate_signals():
        assert not signal.receivers
    assert signal.receivers
    signal.disconnect(_my_callback)


@pytest.mark.skipif(
    django.VERSION >= (4, 0),
    reason='requires `Django<4.0`',
)
@pytest.mark.parametrize('signal', [pre_migrate, post_migrate])
@pytest.mark.usefixtures('migrator', '_disconnect_receivers')
def test_signal_receiver_registered_in_test(mocker, signal):
    """Ensure migration signal receivers registered in tests are called."""
    signal_receiver_mock = mocker.MagicMock()
    main_app_config = apps.get_app_config('main_app')
    signal.connect(
        signal_receiver_mock,
        sender=main_app_config,
        dispatch_uid=DISPATCH_UID,
    )
    verbosity = 0
    interactive = False
    # call `migrate` management command to trigger ``pre_migrate`` and
    # ``post_migrate`` signals
    call_command('migrate', verbosity=verbosity, interactive=interactive)

    signal_receiver_mock.assert_called_once_with(
        sender=main_app_config,
        app_config=main_app_config,
        apps=mocker.ANY,  # we don't have any reference to this object
        using=DEFAULT_DB_ALIAS,
        verbosity=verbosity,
        interactive=interactive,
        plan=mocker.ANY,  # not important for this test
        signal=signal,
    )


@pytest.mark.skipif(
    django.VERSION < (4, 0),
    reason='requires `Django>=4.0`',
)
@pytest.mark.parametrize('signal', [pre_migrate, post_migrate])
@pytest.mark.usefixtures('migrator', '_disconnect_receivers')
def test_signal_receiver_registered_in_test_django40(mocker, signal):
    """Ensure migration signal receivers registered in tests are called."""
    signal_receiver_mock = mocker.MagicMock()
    main_app_config = apps.get_app_config('main_app')
    signal.connect(
        signal_receiver_mock,
        sender=main_app_config,
        dispatch_uid=DISPATCH_UID,
    )
    verbosity = 0
    interactive = False
    # call `migrate` management command to trigger ``pre_migrate`` and
    # ``post_migrate`` signals
    call_command('migrate', verbosity=verbosity, interactive=interactive)

    signal_receiver_mock.assert_called_once_with(
        sender=main_app_config,
        app_config=main_app_config,
        apps=mocker.ANY,  # we don't have any reference to this object
        using=DEFAULT_DB_ALIAS,
        verbosity=verbosity,
        interactive=interactive,
        signal=signal,
        # following kwargs are not important for this test
        stdout=mocker.ANY,
        plan=mocker.ANY,
    )



import subprocess


def test_managepy_check():
    """Checks that checks do fail."""
    proc = subprocess.Popen(
        [
            'python',
            'django_test_app/manage.py',
            'check',
            '--fail-level',
            'WARNING',
        ],
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        encoding='utf8',
    )

    proc.communicate()
    assert proc.returncode == 1



from django_test_migrations.db.backends import exceptions


def test_database_configuration_not_found():
    """Ensure exception returns proper string representation."""
    vendor = 'ms_sql'
    exception = exceptions.DatabaseConfigurationNotFound(vendor)
    assert vendor in str(exception)


def test_database_configuration_setting_not_found():
    """Ensure exception returns proper string representation."""
    vendor = 'ms_sql'
    setting_name = 'fake_setting'
    exception = exceptions.DatabaseConfigurationSettingNotFound(
        vendor,
        setting_name,
    )
    assert vendor in str(exception)
    assert setting_name in str(exception)



import pytest
from django.db.migrations import Migration


@pytest.fixture()
def plan():
    """Fake migrations plan for testing purposes."""
    migrations_plan = [
        Migration('0001_initial', 'app1'),
        Migration('0002_second', 'app1'),
        Migration('0001_initial', 'app2'),
        Migration('0003_third', 'app1'),
        Migration('0004_fourth', 'app1'),
        Migration('0002_second', 'app2'),
        Migration('0005_fifth', 'app1'),
        Migration('0001_initial', 'app3'),
        Migration('0006_sixth', 'app1'),
    ]
    return [(migration, False) for migration in migrations_plan]



import pytest
from django.db.migrations.state import ProjectState

from django_test_migrations.migrator import Migrator


@pytest.mark.django_db()
def test_migrator(transactional_db):
    """We only need this test for coverage."""
    migrator = Migrator()
    old_state = migrator.apply_initial_migration(('main_app', None))
    new_state = migrator.apply_tested_migration(('main_app', '0001_initial'))

    assert isinstance(old_state, ProjectState)
    assert isinstance(new_state, ProjectState)
    assert migrator.reset() is None


@pytest.mark.django_db()
def test_migrator_list(transactional_db):
    """We only need this test for coverage."""
    migrator = Migrator()
    old_state = migrator.apply_initial_migration([('main_app', None)])
    new_state = migrator.apply_tested_migration([('main_app', '0001_initial')])

    assert isinstance(old_state, ProjectState)
    assert isinstance(new_state, ProjectState)
    assert migrator.reset() is None



"""
Django settings for django_test_app project.

Generated by 'django-admin startproject' using Django 2.2.7.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)),
)


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '_dpvr*#hjgv)6v=potf%*+$na7_ck(*+^g08lw0^44zoo88)wb'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    # Our custom checks:
    'django_test_migrations.contrib.django_checks.AutoNames',
    'django_test_migrations.contrib.django_checks.DatabaseConfiguration',

    # Custom:
    'main_app',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'django_test_app.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'django_test_app.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': os.environ.get(
            'DJANGO_DATABASE_ENGINE',
            default='django.db.backends.sqlite3',
        ),
        'USER': os.environ.get('DJANGO_DATABASE_USER', default=''),
        'PASSWORD': os.environ.get('DJANGO_DATABASE_PASSWORD', default=''),
        'NAME': os.environ.get(
            'DJANGO_DATABASE_NAME',
            default=os.path.join(BASE_DIR, 'db.sqlite3'),
        ),
        'PORT': os.environ.get('DJANGO_DATABASE_PORT', default=''),
        'HOST': os.environ.get('DJANGO_DATABASE_HOST', default=''),
    },
}


# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = []


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = '/static/'



import pytest


@pytest.fixture(scope='session')
def django_db_use_migrations(
    request,
    django_db_use_migrations,  # noqa: WPS442
):
    """
    Helper fixture to skip tests when ``--nomigrations`` were specified.

    Copied from https://github.com/pytest-dev/pytest-django
    """
    if not django_db_use_migrations:
        pytest.skip('--nomigrations was specified')
    return django_db_use_migrations



import pytest
from django.core.checks import WARNING

from django_test_migrations.checks.autonames import (
    CHECK_NAME,
    check_migration_names,
)


@pytest.mark.django_db()
def test_autonames():
    """Here we check that bad migrations do produce warnings."""
    warnings = check_migration_names()
    warnings_msgs = {warning.msg for warning in warnings}

    assert len(warnings) == 2

    assert [warnings[0].level, warnings[1].level] == [WARNING, WARNING]
    assert all(
        [
            warnings[0].id.startswith(CHECK_NAME),
            warnings[1].id.startswith(CHECK_NAME),
        ],
    )
    assert warnings_msgs == {
        'Migration main_app.0004_auto_20191119_2125 has an automatic name.',
        'Migration main_app.0005_auto_20200329_1118 has an automatic name.',
    }


@pytest.mark.django_db()
def test_autonames_with_ignore(settings):
    """Here we check that some migrations can be ignored."""
    # patch settings to ignore two bad migrations
    settings.DTM_IGNORED_MIGRATIONS = {
        ('main_app', '0004_auto_20191119_2125'),
        ('main_app', '0005_auto_20200329_1118'),
    }
    warnings = check_migration_names()

    assert not warnings


@pytest.mark.django_db()
def test_autonames_with_ignore_all_app_migrations(settings):
    """Here we check that all migrations ignored inside app."""
    # patch settings to ignore all migrations in the app
    settings.DTM_IGNORED_MIGRATIONS = {('main_app', '*')}
    warnings = check_migration_names()

    assert not warnings



# -*- coding: utf-8 -*-
# Generated by Django 2.2.7 on 2019-11-19 21:25

from django.db import migrations

from main_app.logic.pure.migrations import is_clean_item


def _set_clean_flag(apps, schema_editor):
    """
    Performs the data migration.

    We can't import the ``SomeItem`` model directly as it may be a newer
    version than this migration expects. We use the historical version.

    We are using ``.all()`` because
    we don't have a lot of ``SomeItem`` instances.
    In real-life you should not do that.
    """
    SomeItem = apps.get_model('main_app', 'SomeItem')
    for instance in SomeItem.objects.all():
        instance.is_clean = is_clean_item(instance)
        instance.save(update_fields=['is_clean'])


def _remove_clean_flags(apps, schema_editor):
    """
    This is just a noop example of a rollback function.

    It is not used in our simple case,
    but it should be implemented for more complex scenarios.
    """


class Migration(migrations.Migration):
    """Performs the logical data migration for ``SomeItem``."""

    dependencies = [
        ('main_app', '0002_someitem_is_clean'),
    ]

    operations = [
        migrations.RunPython(_set_clean_flag, _remove_clean_flags),
    ]



from unittest import mock

import django
import pytest
from django.apps import apps
from django.core.management import call_command
from django.db import DEFAULT_DB_ALIAS
from django.db.models.signals import post_migrate, pre_migrate

from django_test_migrations.contrib.unittest_case import MigratorTestCase


class TestSignalMuting(MigratorTestCase):
    """Test that the `post_migrate` signal has been muted."""

    migrate_from = ('main_app', '0002_someitem_is_clean')
    migrate_to = ('main_app', '0001_initial')

    def test_pre_migrate_muted(self):
        """Ensure ``pre_migrate`` signal has been muted."""
        assert not pre_migrate.receivers

    def test_post_migrate_muted(self):
        """Ensure ``post_migrate`` signal has been muted."""
        assert not post_migrate.receivers


class TestSignalConnectInTest(MigratorTestCase):
    """Ensure test ``pre_migrate`` or ``post_migrate`` receiver are called.

    Ensure that ``pre_migrate`` or ``post_migrate`` signals receivers
    connected directly in tests are called.

    """

    migrate_from = ('main_app', '0001_initial')
    migrate_to = ('main_app', '0002_someitem_is_clean')

    def tearDown(self):
        """Disconnect ``pre_migrate`` and ``post_migrate`` testing receivers."""
        pre_migrate.disconnect(
            self.pre_migrate_receiver_mock,
            sender=self.main_app_config,
        )
        post_migrate.disconnect(
            self.post_migrate_receiver_mock,
            sender=self.main_app_config,
        )

    def prepare(self):
        """Connect testing ``pre_migrate`` and ``post_migrate`` receivers."""
        self.pre_migrate_receiver_mock = mock.MagicMock()
        self.post_migrate_receiver_mock = mock.MagicMock()
        # ``old_apps`` is not real ``ProjectState`` instance, so we cannot use
        # it to get "original" main_app ``AppConfig`` instance needed to
        # connect signal receiver, that's the reason we are using
        # ``apps`` imported directly from ``django.apps``
        self.main_app_config = apps.get_app_config('main_app')
        pre_migrate.connect(
            self.pre_migrate_receiver_mock,
            sender=self.main_app_config,
        )
        post_migrate.connect(
            self.post_migrate_receiver_mock,
            sender=self.main_app_config,
        )

    @pytest.mark.skipif(
        django.VERSION >= (4, 0),
        reason='requires `Django<4.0`',
    )
    def test_signal_receivers_added_in_tests(self):
        """Ensure migration signals receivers connected in tests are called."""
        verbosity = 0
        interactive = False
        # call `migrate` management command to trigger ``pre_migrate`` and
        # ``post_migrate`` signals
        call_command('migrate', verbosity=verbosity, interactive=interactive)

        common_kwargs = {
            'sender': self.main_app_config,
            'app_config': self.main_app_config,
            'apps': mock.ANY,  # we don't have any reference to this object
            'using': DEFAULT_DB_ALIAS,
            'verbosity': verbosity,
            'interactive': interactive,
            'plan': mock.ANY,  # not important for this test
        }
        self.pre_migrate_receiver_mock.assert_called_once_with(
            **common_kwargs,
            signal=pre_migrate,
        )
        self.post_migrate_receiver_mock.assert_called_once_with(
            **common_kwargs,
            signal=post_migrate,
        )

    @pytest.mark.skipif(
        django.VERSION < (4, 0),
        reason='requires `Django>=4.0`',
    )
    def test_signal_receivers_added_in_tests_django40(self):
        """Ensure migration signals receivers connected in tests are called."""
        verbosity = 0
        interactive = False
        # call `migrate` management command to trigger ``pre_migrate`` and
        # ``post_migrate`` signals
        call_command('migrate', verbosity=verbosity, interactive=interactive)

        common_kwargs = {
            'sender': self.main_app_config,
            'app_config': self.main_app_config,
            'apps': mock.ANY,  # we don't have any reference to this object
            'using': DEFAULT_DB_ALIAS,
            'verbosity': verbosity,
            'interactive': interactive,
            # following kwargs are not important for this test
            'stdout': mock.ANY,
            'plan': mock.ANY,
        }
        self.pre_migrate_receiver_mock.assert_called_once_with(
            **common_kwargs,
            signal=pre_migrate,
        )
        self.post_migrate_receiver_mock.assert_called_once_with(
            **common_kwargs,
            signal=post_migrate,
        )



from django.db import models

_SomeItemStringFieldLength = 50


class SomeItem(models.Model):
    """We use this model for testing migrations."""

    string_field = models.CharField(max_length=_SomeItemStringFieldLength)
    is_clean = models.BooleanField()



from typing import List, Optional, Tuple, Union

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.migrations import Migration
from django.utils.connection import ConnectionProxy

# Migration target: (app_name, migration_name)
# Regular or rollback migration: 0001 -> 0002, or 0002 -> 0001
# Rollback migration to initial state: 0001 -> None
MigrationTarget = Tuple[str, Optional[str]]
MigrationSpec = Union[MigrationTarget, List[MigrationTarget]]

MigrationPlan = List[Tuple[Migration, bool]]

AnyConnection = Union[ConnectionProxy, BaseDatabaseWrapper]

DatabaseSettingValue = Union[str, int]



"""
This module covers simple direct migrations.

We test both schema and data-migrations here.
"""

import pytest
from django.core.exceptions import FieldError
from django.db.utils import IntegrityError


@pytest.mark.django_db()
def test_pytest_plugin_initial(migrator):
    """Ensures that the initial migration works."""
    old_state = migrator.apply_initial_migration(('main_app', None))

    with pytest.raises(LookupError):
        # Models does not yet exist:
        old_state.apps.get_model('main_app', 'SomeItem')

    new_state = migrator.apply_tested_migration(('main_app', '0001_initial'))
    # After the initial migration is done, we can use the model state:
    SomeItem = new_state.apps.get_model('main_app', 'SomeItem')
    assert SomeItem.objects.filter(string_field='').count() == 0


@pytest.mark.django_db()
def test_pytest_plugin0001(migrator):
    """Ensures that the first migration works."""
    old_state = migrator.apply_initial_migration(('main_app', '0001_initial'))
    SomeItem = old_state.apps.get_model('main_app', 'SomeItem')

    with pytest.raises(FieldError):
        SomeItem.objects.filter(is_clean=True)

    new_state = migrator.apply_tested_migration(
        ('main_app', '0002_someitem_is_clean'),
    )
    SomeItem = new_state.apps.get_model('main_app', 'SomeItem')

    assert SomeItem.objects.filter(is_clean=True).count() == 0


@pytest.mark.django_db()
def test_pytest_plugin0002(migrator):
    """Ensures that the second migration works."""
    old_state = migrator.apply_initial_migration(
        ('main_app', '0002_someitem_is_clean'),
    )
    SomeItem = old_state.apps.get_model('main_app', 'SomeItem')
    SomeItem.objects.create(string_field='a')
    SomeItem.objects.create(string_field='a b')

    assert SomeItem.objects.count() == 2
    assert SomeItem.objects.filter(is_clean=True).count() == 2

    new_state = migrator.apply_tested_migration(
        ('main_app', '0003_update_is_clean'),
    )
    SomeItem = new_state.apps.get_model('main_app', 'SomeItem')

    assert SomeItem.objects.count() == 2
    assert SomeItem.objects.filter(is_clean=True).count() == 1


@pytest.mark.django_db()
def test_pytest_plugin0003(migrator):
    """Ensures that the third migration works."""
    old_state = migrator.apply_initial_migration(
        ('main_app', '0003_update_is_clean'),
    )
    SomeItem = old_state.apps.get_model('main_app', 'SomeItem')
    SomeItem.objects.create(string_field='a')  # default is still there

    assert SomeItem.objects.count() == 1
    assert SomeItem.objects.filter(is_clean=True).count() == 1

    new_state = migrator.apply_tested_migration(
        ('main_app', '0004_auto_20191119_2125'),
    )
    SomeItem = new_state.apps.get_model('main_app', 'SomeItem')

    with pytest.raises(IntegrityError):
        SomeItem.objects.create(string_field='b')  # no default anymore



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""

import os
import sys

from django.core.management import execute_from_command_line

if __name__ == '__main__':
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'django_test_app.settings')
    execute_from_command_line(sys.argv)



import pytest

from django_test_migrations.db.backends import mysql
from django_test_migrations.db.backends.exceptions import (
    DatabaseConfigurationSettingNotFound,
)


def test_get_setting_value(mocker):
    """Ensure expected SQL query is executed."""
    setting_name = 'MAX_EXECUTION_TIME'
    connection_mock = mocker.MagicMock()
    connection_mock.ops.quote_name = lambda name: name
    database_configuration = mysql.configuration.DatabaseConfiguration(
        connection_mock,
    )
    database_configuration.get_setting_value(setting_name)
    cursor_mock = connection_mock.cursor().__enter__()  # noqa: WPS609
    cursor_mock.execute.assert_called_once_with(
        'SELECT @@{0};'.format(setting_name),
    )


def test_get_existing_setting_value(mocker):
    """Ensure setting value is returned for existing setting."""
    expected_setting_value = 74747
    connection_mock = mocker.MagicMock()
    cursor_mock = connection_mock.cursor().__enter__()  # noqa: WPS609
    cursor_mock.fetchone.return_value = (expected_setting_value,)
    database_configuration = mysql.configuration.DatabaseConfiguration(
        connection_mock,
    )
    setting_value = database_configuration.get_setting_value('testing_setting')
    assert setting_value == expected_setting_value


def test_get_not_existing_setting_value(mocker):
    """Ensure exception is raised when setting does not exist."""
    connection_mock = mocker.MagicMock()
    cursor_mock = connection_mock.cursor().__enter__()  # noqa: WPS609
    cursor_mock.fetchone.return_value = None
    database_configuration = mysql.configuration.DatabaseConfiguration(
        connection_mock,
    )
    with pytest.raises(DatabaseConfigurationSettingNotFound):
        database_configuration.get_setting_value('testing_setting')



from django_test_migrations.constants import MIGRATION_TEST_MARKER
from django_test_migrations.contrib.unittest_case import MigratorTestCase


class TestDirectMigration(MigratorTestCase):
    """This class is used to test direct migrations."""

    migrate_from = ('main_app', '0002_someitem_is_clean')
    migrate_to = ('main_app', '0003_update_is_clean')

    def prepare(self):
        """Prepare some data before the migration."""
        SomeItem = self.old_state.apps.get_model('main_app', 'SomeItem')
        SomeItem.objects.create(string_field='a')
        SomeItem.objects.create(string_field='a b')

    def test_migration_main0003(self):
        """Run the test itself."""
        SomeItem = self.new_state.apps.get_model('main_app', 'SomeItem')

        assert SomeItem.objects.count() == 2
        assert SomeItem.objects.filter(is_clean=True).count() == 1


class TestBackwardMigration(MigratorTestCase):
    """This class is used to test backward migrations."""

    migrate_from = ('main_app', '0002_someitem_is_clean')
    migrate_to = ('main_app', '0001_initial')

    def prepare(self):
        """Prepare some data before the migration."""
        SomeItem = self.old_state.apps.get_model('main_app', 'SomeItem')
        SomeItem.objects.create(string_field='a')
        SomeItem.objects.create(string_field='a b')

    def test_migration_main0001(self):
        """Run the test itself."""
        SomeItem = self.new_state.apps.get_model('main_app', 'SomeItem')

        assert SomeItem.objects.count() == 2


def test_migration_test_marker_tag():
    """Ensure ``MigratorTestCase`` subclasses are properly tagged."""
    assert MIGRATION_TEST_MARKER in TestDirectMigration.tags
    assert MIGRATION_TEST_MARKER in TestBackwardMigration.tags



from django.apps import AppConfig
from django.core import checks
from typing_extensions import final

from django_test_migrations.checks import autonames, database_configuration


@final
class AutoNames(AppConfig):
    """
    Class to install this check into ``INSTALLED_APPS`` in ``django``.

    If you have migrations that cannot be renamed,
    use ``DTM_IGNORED_MIGRATIONS`` setting in ``django.conf``
    to ignore ones you have to deal with:

    .. code:: python

        # settings.py
        DTM_IGNORED_MIGRATIONS = {
            ('main_app', '0004_auto_20191119_2125'),
            ('dependency_app', '0001_auto_20201110_2100'),
        }

    To run checks use:

    .. code:: bash

        python manage.py check --deploy --fail-level WARNING

    It will return exit code ``1`` if any violations are found.
    This can be easily added into your CI.

    See:
        https://docs.djangoproject.com/en/3.0/ref/applications/
        https://twitter.com/AdamChainz/status/1231895529686208512

    """

    #: Part of Django API.
    name = autonames.CHECK_NAME

    def ready(self):
        """That's how we register our check when apps are ready."""
        for check in autonames.CHECKS:
            checks.register(check, checks.Tags.compatibility)


@final
class DatabaseConfiguration(AppConfig):
    """Class to install this check into ``INSTALLED_APPS`` in ``django``.

    Database configuration checks are made with aim to help/guide developers
    set the most appropriate values for some database settings according to
    best practices.
    Currently supported database settings:

    * statement timeout (timeout queries that execution take too long):
        * `postgresql` via `statement_timeout` - https://bit.ly/2ZFjaRM
        * `mysql` via `max_execution_time` - https://bit.ly/399TBvk

    See:
        https://github.com/wemake-services/wemake-django-template/issues/1064

    """

    #: Part of Django API.
    name = database_configuration.CHECK_NAME

    def ready(self):
        """Register database configuration checks."""
        for check in database_configuration.CHECKS:
            checks.register(check, checks.Tags.database)



from django_test_migrations.logic.migrations import normalize


def test_normalize_raw_target():
    """Ensure normalize works for ``MigrationTarget``."""
    assert normalize(('app', '0074_magic')) == [('app', '0074_magic')]


def test_normalize_list_of_targets():
    """Ensure normalize works for list of ``MigrationTarget``."""
    migration_targets = [('app1', None), ('app2', '0001_initial')]
    assert normalize(migration_targets) == migration_targets



from typing import ClassVar, Optional

from django.db.migrations.state import ProjectState
from django.db.models.signals import post_migrate, pre_migrate
from django.test import TransactionTestCase, tag

from django_test_migrations.constants import MIGRATION_TEST_MARKER
from django_test_migrations.migrator import Migrator
from django_test_migrations.types import MigrationSpec


@tag(MIGRATION_TEST_MARKER)
class MigratorTestCase(TransactionTestCase):
    """Used when using raw ``unitest`` library for test."""

    database_name: ClassVar[Optional[str]] = None
    old_state: ProjectState
    new_state: ProjectState

    #: Part of the end-user API. Used to tell what migrations we are using.
    migrate_from: ClassVar[MigrationSpec]
    migrate_to: ClassVar[MigrationSpec]

    def setUp(self) -> None:
        """
        Regular ``unittest`` styled setup case.

        What it does?
          - It starts with defining the initial migration state
          - Then it allows to run custom method
            to prepare some data before the migration will happen
          - Then it applies the migration and saves all states

        """
        super().setUp()
        self._migrator = Migrator(self.database_name)
        self.old_state = self._migrator.apply_initial_migration(
            self.migrate_from,
        )
        self.prepare()
        self.new_state = self._migrator.apply_tested_migration(self.migrate_to)

    def prepare(self) -> None:
        """
        Part of the end-user API.

        Used to prepare some data before the migration process.
        """

    def tearDown(self) -> None:
        """Used to clean mess up after each test."""
        pre_migrate.receivers = self._pre_migrate_receivers
        post_migrate.receivers = self._post_migrate_receivers
        self._migrator.reset()
        super().tearDown()

    def _pre_setup(self):
        self._pre_migrate_receivers, pre_migrate.receivers = (  # noqa: WPS414
            pre_migrate.receivers, [],
        )
        self._post_migrate_receivers, post_migrate.receivers = (  # noqa: WPS414
            post_migrate.receivers, [],
        )
        super()._pre_setup()



import pytest

from django_test_migrations.exceptions import MigrationNotInPlan
from django_test_migrations.plan import truncate_plan


@pytest.mark.parametrize(('targets', 'index'), [
    ([], 9),  # full plan for empty targets
    ([('app1', None)], 0),
    ([('app1', None), ('app3', None)], 7),
    ([('app2', '0002_second')], 6),
    ([('app1', '0002_second'), ('app2', None)], 2),
    ([('app1', '0003_third'), ('app2', None)], 4),
    ([('app1', '0003_third'), ('app1', '0005_fifth')], 7),
    ([('app1', '0003_third'), ('app2', None), ('app3', '0001_initial')], 8),
])
def test_truncate_plan(plan, targets, index):
    """Ensure plan is properly truncated for both types migrations names."""
    assert truncate_plan(targets, plan) == plan[:index]


def test_empty_plan():
    """Ensure function work when plan is empty."""
    assert not truncate_plan([('app1', '0001_initial')], [])


@pytest.mark.parametrize('targets', [
    [('app4', None)],
    [('app1', '0047_magic')],
    [('app1', '0005_fifth'), ('app4', None)],
    [('app1', '0005_fifth'), ('app4', '0047_magic'), ('app3', None)],
])
def test_migration_target_does_not_exist(plan, targets):
    """Ensure ``MigrationNotInPlan`` is raised when target not in plan."""
    with pytest.raises(MigrationNotInPlan):
        truncate_plan(targets, plan)



import typing

if typing.TYPE_CHECKING:
    from main_app.models import SomeItem  # noqa: WPS433


def is_clean_item(instance: 'SomeItem') -> bool:
    """Pure function that decides whether or not whitespace is in the model."""
    return ' ' not in instance.string_field



import pytest

from django_test_migrations.plan import all_migrations, nodes_to_tuples


@pytest.mark.django_db()
def test_all_migrations_main():
    """Testing migrations for a single app only."""
    main_migrations = all_migrations('default', ['main_app'])

    assert nodes_to_tuples(main_migrations) == [
        ('main_app', '0001_initial'),
        ('main_app', '0002_someitem_is_clean'),
        ('main_app', '0003_update_is_clean'),
        ('main_app', '0004_auto_20191119_2125'),
        ('main_app', '0005_auto_20200329_1118'),
    ]


@pytest.mark.django_db()
def test_all_migrations_missing():
    """Testing migrations for a missing app."""
    with pytest.raises(LookupError):
        all_migrations('default', ['missing_app'])


@pytest.mark.django_db()
def test_all_migrations_auth():
    """Testing migrations for a builtin app."""
    auth_migrations = all_migrations('default', ['auth'])
    assert len(auth_migrations) >= 10


@pytest.mark.django_db()
def test_all_migrations_all():
    """Testing migrations for all apps."""
    assert len(all_migrations()) >= 17  # noqa: WPS432



import datetime
from typing import List

from django.core.checks import CheckMessage, Warning
from django.db import connections
from typing_extensions import Final

from django_test_migrations.db.backends import exceptions, registry
from django_test_migrations.db.backends.base.configuration import (
    BaseDatabaseConfiguration,
)
from django_test_migrations.logic.datetime import timedelta_to_milliseconds
from django_test_migrations.types import AnyConnection

#: We use this value as a unique identifier of databases related check.
CHECK_NAME: Final = 'django_test_migrations.checks.database_configuration'
STATEMENT_TIMEOUT_MINUTES_UPPER_LIMIT: Final = 30


def check_statement_timeout_setting(*args, **kwargs) -> List[CheckMessage]:
    """Check if statements' timeout settings is properly configured."""
    messages: List[CheckMessage] = []
    for connection in connections.all():
        _check_statement_timeout_setting(connection, messages)
    return messages


def _check_statement_timeout_setting(
    connection: AnyConnection,
    messages: List[CheckMessage],
) -> None:
    try:
        database_configuration = registry.get_database_configuration(
            connection,
        )
    except exceptions.DatabaseConfigurationNotFound:
        return

    try:
        setting_value = int(database_configuration.get_setting_value(
            database_configuration.statement_timeout,
        ))
    except exceptions.DatabaseConfigurationSettingNotFound:
        return

    _ensure_statement_timeout_is_set(
        database_configuration,
        setting_value,
        messages,
    )
    _ensure_statement_timeout_not_too_high(
        database_configuration,
        setting_value,
        messages,
    )


def _ensure_statement_timeout_is_set(
    database_configuration: BaseDatabaseConfiguration,
    setting_value: int,
    messages: List[CheckMessage],
) -> None:
    if not setting_value:
        messages.append(
            Warning(
                '{0}: statement timeout "{1}" setting is not set.'.format(
                    database_configuration.connection.alias,
                    database_configuration.statement_timeout,
                ),
                hint=(
                    'Set "{0}" database setting to some reasonable value.'
                ).format(database_configuration.statement_timeout),
                id='{0}.W001'.format(CHECK_NAME),
            ),
        )


def _ensure_statement_timeout_not_too_high(
    database_configuration: BaseDatabaseConfiguration,
    setting_value: int,
    messages: List[CheckMessage],
) -> None:
    upper_limit = timedelta_to_milliseconds(
        datetime.timedelta(minutes=STATEMENT_TIMEOUT_MINUTES_UPPER_LIMIT),
    )
    if setting_value > upper_limit:
        messages.append(
            Warning(
                (
                    '{0}: statement timeout "{1}" setting value - "{2} ms" ' +
                    'might be too high.'
                ).format(
                    database_configuration.connection.alias,
                    database_configuration.statement_timeout,
                    setting_value,
                ),
                hint=(
                    'Set "{0}" database setting to some ' +
                    'reasonable value, but remember it should not be ' +
                    'too high.'
                ).format(database_configuration.statement_timeout),
                id='{0}.W002'.format(CHECK_NAME),
            ),
        )



# -*- coding: utf-8 -*-
# Generated by Django 2.2.7 on 2019-11-19 21:24

from django.db import migrations, models


class Migration(migrations.Migration):
    """Migration to add ``is_clean`` field to ``SomeItem``."""

    dependencies = [
        ('main_app', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='someitem',
            name='is_clean',
            field=models.BooleanField(default=True),
        ),
    ]



import datetime

import pytest

from django_test_migrations.logic.datetime import timedelta_to_milliseconds


@pytest.mark.parametrize(('timedelta', 'expected_result'), [
    (datetime.timedelta(seconds=1), 1000),
    (datetime.timedelta(minutes=3), 3 * 60 * 1000),
    (datetime.timedelta(hours=2.6), 2.6 * 60 * 60 * 1000),
    (
        datetime.timedelta(days=4),
        4 * 24 * 60 * 60 * 1000,
    ),
    (
        datetime.timedelta(minutes=7.4, seconds=47),
        7.4 * 60 * 1000 + 47 * 1000,
    ),
])
def test_timedelta_to_milliseconds(timedelta, expected_result):
    """Ensure expected value is returned."""
    assert timedelta_to_milliseconds(timedelta) == expected_result



import datetime

import pytest

from django_test_migrations.db.checks.statement_timeout import (
    CHECK_NAME,
    check_statement_timeout_setting,
)
from django_test_migrations.logic.datetime import timedelta_to_milliseconds

ALL_CONNECTIONS_MOCK_PATH = (
    'django_test_migrations.db.checks.statement_timeout.connections.all'
)


@pytest.fixture()
def connection_mock_factory(mocker):
    """Factory of DB connection mocks."""
    def factory(vendor, fetch_one_result=None):
        connection_mock = mocker.MagicMock(vendor=vendor)
        cursor_mock = connection_mock.cursor.return_value
        cursor_mock = cursor_mock.__enter__.return_value  # noqa: WPS609
        cursor_mock.fetchone.return_value = fetch_one_result
        return connection_mock
    return factory


@pytest.mark.parametrize('vendor', ['postgresql', 'mysql'])
def test_correct_statement_timeout(mocker, connection_mock_factory, vendor):
    """Ensure empty list returned when ``statement_timeout`` value correct."""
    connection_mock = connection_mock_factory(vendor, (20000,))
    mocker.patch(ALL_CONNECTIONS_MOCK_PATH, return_value=[connection_mock])
    assert not check_statement_timeout_setting()


@pytest.mark.parametrize('vendor', ['postgresql', 'mysql'])
def test_statement_timeout_not_set(mocker, connection_mock_factory, vendor):
    """Ensure W001 is returned in list when ``statement_timeout`` not set."""
    connection_mock = connection_mock_factory(vendor, (0,))
    mocker.patch(ALL_CONNECTIONS_MOCK_PATH, return_value=[connection_mock])
    check_messages = check_statement_timeout_setting()
    assert len(check_messages) == 1
    assert check_messages[0].id.endswith('W001')


@pytest.mark.parametrize('vendor', ['postgresql', 'mysql'])
def test_statement_timeout_too_high(mocker, connection_mock_factory, vendor):
    """Ensure W002 is returned in list when ``statement_timeout`` too high."""
    connection_mock = connection_mock_factory(
        vendor,
        (timedelta_to_milliseconds(datetime.timedelta(hours=2)),),
    )
    mocker.patch(ALL_CONNECTIONS_MOCK_PATH, return_value=[connection_mock])
    check_messages = check_statement_timeout_setting()
    assert len(check_messages) == 1
    assert check_messages[0].id.endswith('W002')


def test_unsupported_vendors(mocker):
    """Ensure empty list returned when no connections vendors supported."""
    vendors = ['sqlite3', 'custom']
    connection_mocks = [mocker.MagicMock(vendor=vendor) for vendor in vendors]
    mocker.patch(ALL_CONNECTIONS_MOCK_PATH, return_value=connection_mocks)
    assert not check_statement_timeout_setting()


@pytest.mark.parametrize('vendor', ['postgresql', 'mysql'])
def test_statement_timeout_setting_not_found(
    mocker,
    connection_mock_factory,
    vendor,
):
    """Ensure empty list returned when ``statement_timeout`` not found."""
    connection_mock = connection_mock_factory(vendor, None)
    mocker.patch(ALL_CONNECTIONS_MOCK_PATH, return_value=[connection_mock])
    assert not check_statement_timeout_setting()


def test_multiple_connections(mocker, connection_mock_factory):
    """Ensure list with many items returned when many connections present."""
    connections_mocks = [
        connection_mock_factory('sqlite', None),
        connection_mock_factory('postgresql', (0,)),
        connection_mock_factory(
            'mysql',
            (timedelta_to_milliseconds(datetime.timedelta(hours=2)),),
        ),
    ]
    mocker.patch(ALL_CONNECTIONS_MOCK_PATH, return_value=connections_mocks)
    check_messages = check_statement_timeout_setting()
    expected_messages_ids = [
        '{0}.W001'.format(CHECK_NAME),
        '{0}.W002'.format(CHECK_NAME),
    ]
    assert expected_messages_ids == [message.id for message in check_messages]



from typing_extensions import final

from django_test_migrations.db.backends.base.configuration import (
    BaseDatabaseConfiguration,
)
from django_test_migrations.types import DatabaseSettingValue


@final
class DatabaseConfiguration(BaseDatabaseConfiguration):
    """Interact with MySQL database configuration."""

    vendor = 'mysql'
    statement_timeout = 'MAX_EXECUTION_TIME'

    def get_setting_value(self, name: str) -> DatabaseSettingValue:
        """Retrieve value of MySQL database's setting with ``name``."""
        with self.connection.cursor() as cursor:
            cursor.execute(
                'SELECT @@{0};'.format(self.connection.ops.quote_name(name)),
            )
            setting_value = cursor.fetchone()
            if not setting_value:
                return super().get_setting_value(name)
            return setting_value[0]



from typing import Optional

from django.core.management.color import Style, no_style
from django.db import connections
from typing_extensions import Final

DJANGO_MIGRATIONS_TABLE_NAME: Final = 'django_migrations'


def drop_models_tables(
    database_name: str,
    style: Optional[Style] = None,
) -> None:
    """Drop all installed Django's models tables."""
    style = style or no_style()
    connection = connections[database_name]
    tables = connection.introspection.django_table_names(
        only_existing=True,
        include_views=False,
    )
    sql_drop_tables = [
        connection.SchemaEditorClass.sql_delete_table % {
            'table': style.SQL_FIELD(connection.ops.quote_name(table)),
        }
        for table in tables
    ]
    if sql_drop_tables:
        if connection.vendor == 'mysql':
            sql_drop_tables = [
                'SET FOREIGN_KEY_CHECKS = 0;',
                *sql_drop_tables,
                'SET FOREIGN_KEY_CHECKS = 1;',
            ]
        connection.ops.execute_sql_flush(sql_drop_tables)


def flush_django_migrations_table(
    database_name: str,
    style: Optional[Style] = None,
) -> None:
    """Flush `django_migrations` table.

    `django_migrations` is not "regular" Django model, so its not returned
    by ``ConnectionRouter.get_migratable_models`` which is used e.g. to
    implement sequences reset.

    """
    connection = connections[database_name]
    connection.ops.execute_sql_flush(
        connection.ops.sql_flush(
            style or no_style(),
            [DJANGO_MIGRATIONS_TABLE_NAME],
            allow_cascade=False,
            reset_sequences=True,
        ),
    )



from typing import Optional

from django.core.management import call_command
from django.core.management.color import no_style
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.state import ProjectState

from django_test_migrations import sql
from django_test_migrations.logic.migrations import normalize
from django_test_migrations.plan import truncate_plan
from django_test_migrations.signals import mute_migrate_signals
from django_test_migrations.types import MigrationPlan, MigrationSpec


class Migrator(object):
    """
    Class to manage your migrations and app state.

    It is designed to be used inside the tests to ensure that migrations
    are working as intended: both data and schema migrations.

    This class can be but probably should not be used directly.
    Because we have utility test framework
    integrations for ``unitest`` and ``pytest``.

    Use them for better experience.
    """

    def __init__(
        self,
        database: Optional[str] = None,
    ) -> None:
        """That's where we initialize all required internals."""
        if database is None:
            database = DEFAULT_DB_ALIAS

        self._database: str = database
        self._executor = MigrationExecutor(connections[self._database])

    def apply_initial_migration(self, targets: MigrationSpec) -> ProjectState:
        """Reverse back to the original migration."""
        targets = normalize(targets)

        style = no_style()
        # start from clean database state
        sql.drop_models_tables(self._database, style)
        sql.flush_django_migrations_table(self._database, style)

        # prepare as broad plan as possible based on full plan
        self._executor.loader.build_graph()  # reload
        full_plan = self._executor.migration_plan(
            self._executor.loader.graph.leaf_nodes(),
            clean_start=True,
        )
        plan = truncate_plan(targets, full_plan)

        # apply all migrations from generated plan on clean database
        # (only forward, so any unexpected migration won't be applied)
        # to restore database state before tested migration
        return self._migrate(targets, plan=plan)

    def apply_tested_migration(self, targets: MigrationSpec) -> ProjectState:
        """Apply the next migration."""
        self._executor.loader.build_graph()  # reload
        return self._migrate(normalize(targets))

    def reset(self) -> None:
        """
        Reset the state to the most recent one.

        Notably, signals are not muted here to avoid
        https://github.com/wemake-services/django-test-migrations/issues/128

        """
        call_command('migrate', verbosity=0, database=self._database)

    def _migrate(
        self,
        migration_targets: MigrationSpec,
        plan: Optional[MigrationPlan] = None,
    ) -> ProjectState:
        with mute_migrate_signals():
            return self._executor.migrate(migration_targets, plan=plan)



import pytest

from django_test_migrations.db.backends import mysql, postgresql, registry
from django_test_migrations.db.backends.base.configuration import (
    BaseDatabaseConfiguration,
)
from django_test_migrations.db.backends.exceptions import (
    DatabaseConfigurationNotFound,
)


def test_all_db_backends_registered():
    """Ensures all database backends all registered."""
    registered_vendors = list(registry.database_configuration_registry.keys())
    assert sorted(registered_vendors) == ['mysql', 'postgresql']


def test_abc_subclasses_are_not_registered():
    """Test registration of ``BaseDatabaseConfiguration`` abstract subclasses.

    Ensures ``BaseDatabaseConfiguration`` abstract subclasses are not
    registered.
    """
    vendor = 'abstract_subclass'
    # creates abstract subclass
    type('DatabaseConfiguration', (BaseDatabaseConfiguration,), {
        'vendor': vendor,
    })
    assert vendor not in registry.database_configuration_registry


@pytest.mark.parametrize(('vendor', 'database_configuration_class'), [
    ('postgresql', postgresql.configuration.DatabaseConfiguration),
    ('mysql', mysql.configuration.DatabaseConfiguration),
])
def test_get_database_configuration_vendor_registered(
    mocker,
    vendor,
    database_configuration_class,
):
    """Ensures database configuration is returned when vendor registered."""
    connection_mock = mocker.Mock()
    connection_mock.vendor = vendor
    database_configuration = registry.get_database_configuration(
        connection_mock,
    )
    assert isinstance(database_configuration, database_configuration_class)


def test_get_database_configuration_vendor_not_registered(mocker):
    """Ensures proper exception is raised when vendor not registered."""
    vendor = 'not_registered_vendor'
    connection_mock = mocker.Mock()
    connection_mock.vendor = vendor
    with pytest.raises(DatabaseConfigurationNotFound, match=vendor):
        registry.get_database_configuration(connection_mock)



# -*- coding: utf-8 -*-
# Generated by Django 2.2.7 on 2019-11-19 21:25

"""
This migration is named incorrectly.

We use it as a test for wrong autonames.
Please, do not rename it!
"""

from django.db import migrations, models


class Migration(migrations.Migration):
    """Removes the default value from ``is_clean`` from ``SomeItem``."""

    dependencies = [
        ('main_app', '0003_update_is_clean'),
    ]

    operations = [
        migrations.AlterField(
            model_name='someitem',
            name='is_clean',
            field=models.BooleanField(),
        ),
    ]



import re
import subprocess

from django_test_migrations.constants import MIGRATION_TEST_MARKER


def test_call_pytest_setup_plan():
    """Checks that module is registered and visible in the meta data."""
    output_text = subprocess.check_output(
        [
            'pytest',
            '--setup-plan',

            # We need this part because otherwise check fails with `1` code:
            '--cov-fail-under',
            '0',
        ],
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        encoding='utf8',
    )

    assert 'migrator' in output_text
    assert 'migrator_factory' in output_text


def test_pytest_registers_marker():
    """Ensure ``MIGRATION_TEST_MARKER`` marker is registered."""
    output_text = subprocess.check_output(
        ['pytest', '--markers'],
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        encoding='utf8',
    )

    assert MIGRATION_TEST_MARKER in output_text


def test_pytest_markers():
    """Ensure ``MIGRATION_TEST_MARKER`` markers are properly added."""
    output_text = subprocess.check_output(
        [
            'pytest',
            '--collect-only',

            # Collect only tests marked with ``MIGRATION_TEST_MARKER`` marker
            '-m',
            MIGRATION_TEST_MARKER,

            # We need this part because otherwise check fails with `1` code:
            '--cov-fail-under',
            '0',
        ],
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        encoding='utf8',
    )

    search_result = re.search(
        r'(?P<selected_number>\d+)\s+selected',
        output_text,
    )
    assert search_result
    assert int(search_result.group('selected_number') or 0) > 0
    assert 'test_pytest_plugin' in output_text



from django_test_migrations.sql import drop_models_tables

TESTING_DATABASE_NAME = 'test'


def test_drop_models_table_no_tables_detected(mocker):
    """Ensure any ``DROP TABLE`` statement executed when no tables detected."""
    testing_connection_mock = mocker.MagicMock()
    testing_connection_mock.introspection.django_table_names.return_value = []
    connections_mock = mocker.patch('django.db.connections._connections')
    connections_mock.test = testing_connection_mock

    drop_models_tables(TESTING_DATABASE_NAME)

    testing_connection_mock.ops.execute_sql_flush.assert_not_called()


def test_drop_models_table_table_detected(mocker):
    """Ensure ``DROP TABLE`` statements are executed when any table detected."""
    testing_connection_mock = mocker.MagicMock()
    testing_connection_mock.introspection.django_table_names.return_value = [
        'foo_bar',
        'foo_baz',
    ]
    connections_mock = mocker.patch('django.db.connections._connections')
    connections_mock.test = testing_connection_mock

    drop_models_tables(TESTING_DATABASE_NAME)

    testing_connection_mock.ops.execute_sql_flush.assert_called_once()


def test_drop_models_table_on_mysql(mocker):
    """Ensure queries disabling/enabling `FOREIGN_KEY_CHECKS` are executed."""
    testing_connection_mock = mocker.MagicMock(vendor='mysql')
    testing_connection_mock.introspection.django_table_names.return_value = [
        'foo_bar',
        'foo_baz',
    ]
    connections_mock = mocker.patch('django.db.connections._connections')
    connections_mock.test = testing_connection_mock

    drop_models_tables(TESTING_DATABASE_NAME)

    testing_connection_mock.ops.execute_sql_flush.assert_called_once_with([
        'SET FOREIGN_KEY_CHECKS = 0;',
        mocker.ANY,
        mocker.ANY,
        'SET FOREIGN_KEY_CHECKS = 1;',
    ])



# register all ``BaseDatabaseConfiguration`` subclasses
from django_test_migrations.db.backends.mysql.configuration import (
    DatabaseConfiguration as MySQLDatabaseConfiguration,
)
from django_test_migrations.db.backends.postgresql.configuration import (
    DatabaseConfiguration as PostgreSQLDatabaseConfiguration,
)



# -*- coding: utf-8 -*-
# Generated by Django 2.2.10 on 2020-03-29 11:18

"""
This migration is named incorrectly.

We use it as a test for wrong autonames.
Please, do not rename it!
"""

from django.db import migrations


class Migration(migrations.Migration):
    """Dummy migration."""

    dependencies = [
        ('main_app', '0004_auto_20191119_2125'),
    ]

    operations = []



from contextlib import contextmanager

from django.db.models.signals import post_migrate, pre_migrate


@contextmanager
def mute_migrate_signals():
    """Context manager to mute migration-related signals."""
    pre_migrate_receivers = pre_migrate.receivers
    post_migrate_receivers = post_migrate.receivers
    pre_migrate.receivers = []
    post_migrate.receivers = []
    yield pre_migrate_receivers, post_migrate_receivers
    pre_migrate.receivers = pre_migrate_receivers
    post_migrate.receivers = post_migrate_receivers



"""
This module covers tests for migration rollbacks.

It might be useful when something goes wrong
and you need to switch back to the previous state.
"""

import pytest
from django.core.exceptions import FieldError


@pytest.mark.django_db()
def test_pytest_plugin0001(migrator):
    """Ensures that the first migration works."""
    old_state = migrator.apply_initial_migration(
        ('main_app', '0002_someitem_is_clean'),
    )
    SomeItem = old_state.apps.get_model('main_app', 'SomeItem')

    assert SomeItem.objects.filter(is_clean=True).count() == 0

    new_state = migrator.apply_tested_migration(('main_app', '0001_initial'))
    SomeItem = new_state.apps.get_model('main_app', 'SomeItem')

    with pytest.raises(FieldError):
        SomeItem.objects.filter(is_clean=True)


@pytest.mark.django_db()
def test_pytest_plugin0002(migrator):
    """Ensures that the second migration works."""
    old_state = migrator.apply_initial_migration(
        ('main_app', '0003_update_is_clean'),
    )
    SomeItem = old_state.apps.get_model('main_app', 'SomeItem')
    SomeItem.objects.create(string_field='a', is_clean=True)
    SomeItem.objects.create(string_field='a b', is_clean=False)

    assert SomeItem.objects.count() == 2
    assert SomeItem.objects.filter(is_clean=True).count() == 1

    new_state = migrator.apply_tested_migration(
        ('main_app', '0002_someitem_is_clean'),
    )
    SomeItem = new_state.apps.get_model('main_app', 'SomeItem')

    assert SomeItem.objects.count() == 2
    assert SomeItem.objects.filter(is_clean=True).count() == 1



import pytest
from django.core.management.color import Style

from django_test_migrations import sql


@pytest.fixture()
def testing_connection_mock(mocker):
    """Mock Django connections to check the methods called."""
    testing_connection_mock = mocker.MagicMock()
    testing_connection_mock.introspection.get_sequences.return_value = []
    connections_mock = mocker.patch('django.db.connections._connections')
    connections_mock.test = testing_connection_mock
    return testing_connection_mock


def test_flush_django_migration_table(mocker, testing_connection_mock):
    """Ensure expected ``connection`` methods are called."""
    style = Style()

    sql.flush_django_migrations_table('test', style)

    testing_connection_mock.ops.sql_flush.assert_called_once_with(
        style,
        [sql.DJANGO_MIGRATIONS_TABLE_NAME],
        reset_sequences=True,
        allow_cascade=False,
    )
    testing_connection_mock.ops.execute_sql_flush.assert_called_once_with(
        mocker.ANY,
    )



import pytest

from django_test_migrations import exceptions


@pytest.mark.parametrize(('target', 'expected_str'), [
    (('app', None), "Migration ('app', None) not found in migrations plan"),
    (
        ('app', '0047_magic'),
        "Migration ('app', '0047_magic') not found in migrations plan",
    ),
])
def test_representation(target, expected_str):
    """Ensure ``MigrationNotInPlan`` has expected string representation."""
    assert str(exceptions.MigrationNotInPlan(target)) == expected_str
